/*****************************************************************************
 * Copyright © 2015, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
 ****************************************************************************/

import groovy.transform.Memoized

import org.hyperic.sigar.Sigar
import org.hyperic.sigar.ptql.ProcessFinder
import org.apache.commons.logging.Log
import org.apache.commons.logging.LogFactory
import org.hyperic.hq.hqapi1.types.Agent
import org.hyperic.hq.hqapi1.types.AlertDefinitionsResponse
import org.hyperic.hq.hqapi1.types.ControlHistory
import org.hyperic.hq.hqapi1.types.EscalationsResponse
import org.hyperic.hq.hqapi1.types.GroupsResponse
import org.hyperic.hq.hqapi1.types.MetricTemplate
import org.hyperic.hq.hqapi1.types.MetricTemplatesResponse
import org.hyperic.hq.hqapi1.types.PingAgentResponse
import org.hyperic.hq.hqapi1.types.Response
import org.hyperic.hq.hqapi1.types.ResponseStatus
import org.hyperic.hq.hqapi1.types.Resource
import org.hyperic.hq.hqapi1.types.ResourceConfig
import org.hyperic.hq.hqapi1.types.ResourcePrototype
import org.hyperic.hq.hqapi1.types.ResourceResponse
import org.hyperic.hq.hqapi1.types.StatusResponse
import org.hyperic.hq.hqapi1.HQApi
import org.hyperic.hq.hqapi1.ResourceApi
import org.hyperic.hq.hqapi1.ControlApi
import org.hyperic.hq.hqapi1.XmlUtil

import javax.xml.bind.JAXBContext
import javax.xml.bind.JAXBElement
import javax.xml.bind.Marshaller
import javax.xml.namespace.QName

/**
 * Convenience methods for interacting with SAS Environment Manager via
 * the Hyperic HQAPI
 */
class HQApiUtil
{
    // Get default (file) logger
    private Log log = LogFactory.getLog(HQApiUtil.class)
    // Base script provides helpful methods (required in constructor)
    private EMIScript emi

    // Some methods (sync*) return a simple true/false
    // Detailed status is available in a separate call
    ResponseStatus lastStatus

    // Default base directory to the working directory, but allow scripts to override
    String baseDir = "."

    static {
        // Implement asBoolean() for HQApi ResponseStatus and Response
        // for easy use in asserts, if statements, etc.
        ResponseStatus.metaClass.asBoolean = {
            delegate == ResponseStatus.SUCCESS
        }
        Response.metaClass.asBoolean = {
            delegate.status == ResponseStatus.SUCCESS
        }
    }

    // Constructor requires a base script
    // Only should be called by infrastructure, not directly by scripts
    HQApiUtil(EMIScript base) {
        if (!base) {
            // If this ever happens, it is a coding error
            throw new RuntimeException("Missing base script")
        }
        emi = base
    }

    /**
     * Returns an instance of Sigar to work with
     *
     * @return an instance of Sigar
     */
    @Memoized
    Sigar getSigar() {
        return new Sigar()
    }

    /**
     * Retrieves a configured HQApi instance
     *
     * @return an HQApi instance with the ability to query the same server the local
     *         SAS Environment Manager Agent is reporting to
     */
    @Memoized
    HQApi getApi() {
        def connInfo = getConnectionInfo()
        new HQApi(connInfo['hostname'], connInfo['port'], connInfo['isSecure'],
                  connInfo['user'], connInfo['password'])
    }

    /**
     * Returns information on the Agent for the local host
     *
     * @return the Agent for the local host
     */
    Agent getAgent() {
        for (Agent agent : api.agentApi.agents.agent) {
            if (emi.net.isLocal(agent.address)) {
                return agent
            }
        }
        log.warn "Unable to find registered Agent for local host"
        return null
    }

    /**
     * Returns the base directory location of the SAS Environment Manager Agent
     *
     * @return the base directory location of the SAS Environment Manager Agent
     */
    @Memoized
    File getAgentDir() {
        return emi.canonicalFile(System.getProperty("sasev.agent.dir"))
    }

    /**
     * Returns the contents of the conf/agent.properties configuration file
     *
     * @return the contents of the Agent properties file
     */
    @Memoized
    Properties getAgentProperties() {
        return emi.readProperties(new File(System.getProperty("sasev.agent.prop.file")))
    }

    /**
     * Returns true if the SAS Environment Manager Agent process is running. Note that 'running'
     * does not necessarily indicate that the Agent is successfully connected to the server and
     * reporting metrics.
     *
     * @return True if the SASEV Agent process is running, false otherwise
     */
    boolean isAgentRunning() {
        boolean status = false
        try {
            String agentSBin = emi.canonicalFile(emi.sysProps['sasev.agent.dir'], "wrapper/sbin")
            String agentQuery = "State.Name.eq=java,Exe.Cwd.eq=$agentSBin"
            log.debug "SASEV Agent PTQL query is $agentQuery"
            long[] pids = ProcessFinder.find(sigar, agentQuery)
            if (pids.length == 0) {
                log.debug "OS Version [${System.properties['os.name']}]"
                def agentPIDFile = emi.canonicalFile(emi.sysProps['sasev.agent.dir'],"wrapper/hq-agent.pid")
                if (agentPIDFile.exists() && System.properties['os.name'].toUpperCase().contains('AIX') ) {
                  status = true
                }
                else {
                  log.debug "No SASEV Agent process running in $agentSBin"
                }
            } else if (pids.length == 1) {
                log.debug "Found SASEV Agent at PID ${pids[0]}"
                status = true
            } else {
                log.debug "Multiple agent processes matched query. Returning true."
                status = true
            }
        } catch (Exception e) {
            log.error "Unable to determine agent state. Assuming DOWN.", e
        }

        log.debug "SASEV Agent status is $status"
        return status
    }

    /**
     * Returns true if the SAS Environment Manager Agent process is running and communicating
     * with the server.
     *
     * @return True if the SASEV Agent process is running, false otherwise
     */
    boolean isAgentAvailable() {
        // S1395976 - Call to agentPing removed as it was unreliable
        return agentRunning
    }

    /**
     * Sets the credentials used for connecting to the SAS Environment Manager Server
     *
     * @param username
     * @param password
     */
    void setCredentials(String username, String password) {
        def connInfo = getConnectionInfo()
        api = new HQApi(connInfo['hostname'], connInfo['port'], connInfo['isSecure'], username, password)
    }

    /**
     * Getter for the read-only lastStatus property
     *
     * @return The status of the last sync operation or null if no status was available
     */
    ResponseStatus getLastStatus() {
        return lastStatus
    }

    /**
     * Resolves the given filename against the current base directory if not already absolute
     *
     * @param filename
     *
     * @return A File object to the canonical path to the filename resovled against the
     *         current base directory
     */
    File resolveFile(String filename) {
        File file = new File(filename)
        if (!file.isAbsolute()) {
            file = new File(baseDir, filename)
        }
        try {
            file = file.canonicalFile
        } catch (IOException ioEx) {
            // Ignore...we'll just deal with the absolute file
        }

        return file
    }

    /**
     * Converts any HQApi object to its XML representation
     *
     * @param o The HQApi object to marshall to XML
     *
     * @return A String containing the XML representation of the object
     */
     String toXML(Object o) {
         JAXBContext jaxbContext = JAXBContext.newInstance(o.getClass())
         Marshaller jaxbMarshaller = jaxbContext.createMarshaller()

         // format the XML output
         jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true)
         String[] c = o.getClass().name.split('\\.')
         QName qName = new QName(null, c[c.size()-1])
         JAXBElement root = new JAXBElement(qName, o.getClass(), o)

         StringWriter stringWriter = new StringWriter()
         jaxbMarshaller.marshal(root, stringWriter)

         String xml = stringWriter.toString()
         return xml
     }

    /**
     * Syncs the given path as metric definitions XML
     *
     * @param filename The path to the metric definition XML file
     *
     * @return true if successful, false otherwise. Detailed failure status
     *         is available via {@link #getLastStatus()}
     */
    boolean syncMetricXML(String filename) {
        syncMetricXML resolveFile(filename)
    }

    /**
    * Syncs the given file as metric definitions XML
    *
    * @param file The metric definition XML File
    *
    * @return true if successful, false otherwise. Detailed failure status
    *         is available via {@link #getLastStatus()}
    */
    boolean syncMetricXML(File file) {
        log.info "Syncing XML metrics file [${file.name}]..."
        try {
            def syncStatus = api.metricApi.syncMetricTemplates readMetricXML(file)
            if (syncStatus.status.equals(ResponseStatus.SUCCESS)) {
                log.info "Metric template sync successful for [${file.name}]"
            } else {
                log.warn "Metric template sync unsuccessful for [${file.name}] : ${syncStatus?.error?.errorCode} ${syncStatus.error.reasonText}"
            }

            lastStatus = syncStatus.status
        } catch (Exception e) {
            log.error "There was a problem syncing metric XML", e
            lastStatus = ResponseStatus.FAILURE
        }

        return lastStatus.equals(ResponseStatus.SUCCESS)
    }

    /**
    * Syncs the given path as group definitions XML
    *
    * @param filename The path to the group definition XML file
    *
    * @return true if successful, false otherwise. Detailed failure status
    *         is available via {@link #getLastStatus()}
    */
    boolean syncGroupXML(String filename) {
        syncGroupXML resolveFile(filename)
    }

    /**
    * Syncs the given file as group definitions XML
    *
    * @param file The group definition XML File
    *
    * @return true if successful, false otherwise. Detailed failure status
    *         is available via {@link #getLastStatus()}
    */
    boolean syncGroupXML(File file) {
        log.info "Syncing XML group definitions file [${file.name}]..."
        // The alert definition sync API is strange in that it doesn't return a StatusResponse
        def syncStatus = api.alertDefinitionApi.syncAlertDefinitions readAlertsXML(file)
        if (syncStatus) {
            log.info "Group Definition sync successful for [${file.name}]"
        } else {
            log.warn "Group Definition sync unsuccessful for [${file.name}] : ${syncStatus?.error?.errorCode} ${syncStatus.error.reasonText}"
        }

        lastStatus = null
        return syncStatus != null
    }

    /**
    * Syncs the given path as alert definitions XML
    *
    * @param filename The path to the alert definition XML file
    *
    * @return true if successful, false otherwise. Detailed failure status
    *         is available via {@link #getLastStatus()}
    */
    boolean syncAlertXML(String filename) {
        syncAlertXML resolveFile(filename)
    }

    /**
    * Syncs the given file as alert definitions XML
    *
    * @param file The alert definition XML File
    *
    * @return true if successful, false otherwise. Detailed failure status
    *         is available via {@link #getLastStatus()}
    */
    boolean syncAlertXML(File file) {
        log.info "Syncing XML alert definitions file [${file.name}]..."
        // The alert definition sync API is strange in that it doesn't return a StatusResponse
        // Returns an AlertDefinitionResponse with a 'status' attribute
        def syncStatus = api.alertDefinitionApi.syncAlertDefinitions readAlertsXML(file)
        log.debug "Alert sync status: ${syncStatus?.dump()}"
        if (syncStatus.status) {
            log.info "Alert Definition sync successful for [${file.name}]"
        } else {
            log.warn "Alert Definition sync unsuccessful for [${file.name}] : ${syncStatus?.error?.errorCode} ${syncStatus.error.reasonText}"
        }

        lastStatus = syncStatus?.status
        return syncStatus as boolean
    }

    /**
    * Syncs the given path as escalation XML
    *
    * @param filename The path to the escaltion XML file
    *
    * @return true if successful, false otherwise. Detailed failure status
    *         is available via {@link #getLastStatus()}
    */
    boolean syncEscalationXML(String filename) {
        syncEscalationXML resolveFile(filename)
    }

    /**
    * Syncs the given file as escalation XML
    *
    * @param file The escaltion XML File
    *
    * @return true if successful, false otherwise. Detailed failure status
    *         is available via {@link #getLastStatus()}
    */
    boolean syncEscalationXML(File file) {
        log.info "Syncing XML escalations file [${file.name}]..."
        def syncStatus = api.escalationApi.syncEscalations readEscalationsXML(file)
        if (syncStatus.status.equals(ResponseStatus.SUCCESS)) {
            log.info "Escalation sync successful for [${file.name}]"
        } else {
            log.warn "Escalation sync unsuccessful for [${file.name}] : ${syncStatus?.error?.errorCode} ${syncStatus.error.reasonText}"
        }

        lastStatus = syncStatus.status
        return syncStatus.status.equals(ResponseStatus.SUCCESS)
    }

    /**
    * Syncs the given path as metric template XML
    *
    * @param filename The path to the metric template XML file
    *
    * @return true if successful, false otherwise. Detailed failure status
    *         is available via {@link #getLastStatus()}
    */
    boolean syncMetricTemplates(String filename) {
        syncMetricTemplates new File(filename)
    }

    /**
    * Syncs the given file as metric template XML
    *
    * @param filename The metric template XML File
    *
    * @return true if successful, false otherwise. Detailed failure status
    *         is available via {@link #getLastStatus()}
    */
    boolean syncMetricTemplates(File metricFile) {
        metricFile = resolveFile(metricFile.path)
        // =============================
        // Sync metric template XML file
        // =============================
        // name:ResourcePrototype
        Map protoMap = [:]
        // name:MetricTemplate list
        Map templateMap = [:]
        Set modifiedTemplates = [] as HashSet

        // Lines are expected to be ResourcePrototypeName|MetricAlias|Attribute=value
        String lineRegEx = /(.+)[|](.+)[|](.+)=(.+)/
        log.info "Loading $metricFile..."
        metricFile.eachLine { line ->
            line = line.trim()
            def m = (line =~ lineRegEx)
            if (m.matches()) {
                String resType = m[0][1]
                String metricName = m[0][2]
                String attr = m[0][3]
                String value = m[0][4]

                ResourcePrototype proto = protoMap[resType]
                if (!proto) {
                    log.info "Finding resource prototype [$resType]..."
                    proto = api.resourceApi.getResourcePrototype(resType).resourcePrototype
                    protoMap[resType] = proto
                    templateMap[resType] = api.metricApi.getMetricTemplates(proto).metricTemplate
                }

                MetricTemplate template = findTemplateByAlias(templateMap[resType], metricName)
                if (template) {
                    // Update the template
                    log.info "Setting [$resType]:[$metricName]:[$attr] to [$value]"
                    switch (attr) {
                        case "defaultInterval":
                            template[attr] = Long.parseLong(value)
                            break
                        case ["defaultOn","isIndicator"]:
                            template[attr] = Boolean.valueOf(value)
                            break
                        default:
                            template[attr] = value
                    }
                    modifiedTemplates.add(template)
                } else {
                    log.warn "WARNING: The metric alias [$metricName] does not exist for [$resType]"
                }
            } else {
                log.info "The line did not match the pattern\n\t$line"
            }
        }

        log.info "Syncing ${modifiedTemplates.size()} templates"

        if (modifiedTemplates) {
            StatusResponse syncStatus = api.metricApi.syncMetricTemplates(modifiedTemplates.asList())
            if (syncStatus.status.equals(ResponseStatus.SUCCESS)) {
                log.info "Metric Template sync successful!"
            } else {
                log.warn "Metric Template sync unsuccessful"
            }
            lastStatus = syncStatus.status
            return syncStatus.status.equals(ResponseStatus.SUCCESS)
        }

        lastStatus = null
        return true
    }

    /**
     * Finds a MetricTemplate from the provided List with the given alias
     *
     * @param templateList A List of MetricTemplates
     * @param alias The alias of the MetricTemplate to find
     *
     * @return the first MetricTemplate in the List with the given alias
     */
    def MetricTemplate findTemplateByAlias(def templateList, def alias) {
        for (MetricTemplate template : templateList) {
            if (template.alias == alias) {
                return template
            }
        }

        return null
    }

    /**
     * Converts the given XML file to a MetricTemplatesResponse object
     *
     * @param xmlFile File to deserialize
     *
     * @return a MetricTemplatesResponse object deserialized from the XML
     */
    def readMetricXML(File xmlFile) {
        xmlFile.withInputStream { inStream ->
            MetricTemplatesResponse apiResp = XmlUtil.deserialize(MetricTemplatesResponse.class, inStream)
            return apiResp.metricTemplate
        }
    }

    /**
    * Converts the given XML file to a GroupsResponse object
    *
    * @param xmlFile File to deserialize
    *
    * @return a GroupsResponse object deserialized from the XML
    */
    def readGroupXML(File xmlFile) {
        xmlFile.withInputStream { inStream ->
            GroupsResponse apiResp = XmlUtil.deserialize(GroupsResponse.class, inStream)
            return apiResp.group
        }
    }

    /**
    * Converts the given XML file to a AlertDefinitionsResponse object
    *
    * @param xmlFile File to deserialize
    *
    * @return a AlertDefinitionsResponse object deserialized from the XML
    */
    def readAlertsXML(File xmlFile) {
        xmlFile.withInputStream { inStream ->
            AlertDefinitionsResponse apiResp = XmlUtil.deserialize(AlertDefinitionsResponse.class, inStream)
            return apiResp.alertDefinition
        }
    }

    /**
    * Converts the given XML file to a EscalationsResponse object
    *
    * @param xmlFile File to deserialize
    *
    * @return a EscalationsResponse object deserialized from the XML
    */
    def readEscalationsXML(File xmlFile) {
        xmlFile.withInputStream { inStream ->
            EscalationsResponse apiResp = XmlUtil.deserialize(EscalationsResponse.class, inStream)
            return apiResp.escalation
        }
    }

    /**
     * Creates a platform service with the given parameters. If a service of the same name, type, and
     * parent already exists, this method will do nothing unless forceDelete==true
     *
     * @param type The name of the resource type of the platform service. This parameter cannot be null.
     * @param platform The FQDN of the platform that will host the platform service. The parent
     *                 will default to the local machine if this parameter is null
     * @param name The name of the platform service instance. This parameter cannot be null.
     * @param forceDelete Delete existing platform service if it exists
     * @param config Configuration options for the platform service
     *
     * @return The ServiceResource representing the platform service or null if it could not be created
     */
    def createPlatformService(String type, String platform, String name, boolean forceDelete, Map config=[:]) {
        Resource svc = null

        // Check parameters
        if (!type) {
            throw new RuntimeException("Missing resource type:[$type]")
        }
        if (!name) {
            throw new RuntimeException("Missing resource name:[$name]")
        }
        if (!platform) {
            platform = emi.net.fqdn
        }

        // Get ResourcePrototype
        ResourceApi resApi = api.resourceApi
        Resource platformResource = resApi.getPlatformResourceByFqdn(platform, false, false)?.resource
        def proto = resApi.getResourcePrototype(type)?.resourcePrototype

        if (!platformResource) {
            log.error "Cannot create [$name]. Platform $platform missing."
            return null
        }

        if (!proto) {
            log.error "Cannot create [$name]. Resource type $type missing."
            return null
        }

        log.debug "Checking if [$type:$name] exists..."
        // Do not create if the platform service already exists (same name, same type, same platform)
        resApi.getResources(proto, true, false).resource.each {
            if (it.name == name) {
                Resource parentServer = resApi.getParent(it).resource
                Resource parentPlatform = resApi.getParent(parentServer).resource
                if (platform == parentPlatform?.name) {
                    svc = it
                    return // from closure
                }
            }
        }

        boolean resourceExists = (svc != null)
        if (resourceExists && forceDelete) {
            log.debug "The platform service [$name] already exists. Forcing a delete."
            def deleteResp = resApi.deleteResource(svc.id)
            if (deleteResp) {
                resourceExists = false
                svc = null
            } else {
                log.warn "WARNING: Unable to delete platform service [$name]"
            }
        }

        if (resourceExists) {
            log.info "The platform service [$name] already exists. Skipping create."
        } else {
            // Create the platform service
            log.debug "\nCreating platform service [\n\t$type\n\t$platform\n\t$name\n\t$config\n]"
            ResourceResponse resResp = resApi.createService(proto, platformResource, name, config)
            lastStatus = resResp.status
            if (resResp) {
                svc = resResp.resource
                // Double=check that configuration was set - look for hostname which should always be there
                if (config['hostname'] != null) {
                    boolean verifiedConfig = false
                    for (ResourceConfig resCfg : svc.resourceConfig) {
                        if  (resCfg.key == "hostname") {
                            if (resCfg.value == config['hostname']) {
                                verifiedConfig = true
                                break
                            } else {
                                log.warn "Platform service [$name] was not created properly"
                                log.warn "$resCfg.value does not match expected value $config['hostname']"
                                break
                            }
                        }
                    }
                    if (!verifiedConfig) {
                        log.warn "Deleting partially created resource [$name]..."
                        def deleteResp = resApi.deleteResource(svc.id)
                        if (!deleteResp) {
                            log.warn "[$name] could not be deleted. Check the resource in SAS Environment " +
                                     "Manager and manually delete if necessary."
                        }
                    }
                }
            } else {
                log.warn "WARNING: Unable to create platform service ($type, $platform, $name, $config): ${resResp?.error?.reasonText}"
                // Delete any partially-created resources
                if (resResp?.resource) {
                    resResp.resource.each { resource ->
                        try {
                            resApi.deleteResource(resource.id)
                        } catch (Exception e) {
                            log.warn "WARNING: Unable to delete partially created resource [${resource?.dump()}]: $e"
                        }
                    }
                }
            }
        }

        return svc
    }

    /**
     * Runs the specified control action on the given resource with the
     * given (optional) arguments. This method is synchronous and will wait
     * for the control action to complete.
     *
     * @param r The control action will run against this resource instance
     * @param action The control action to run
     * @param args Optional arguments to pass to the control action invocation
     *
     * @return The ControlHistory for the completed control action. This
     *         object can be inspected for success/failure status.
     */
    ControlHistory runControlAction(Resource r, String action, String[] args = [] as String[]) {
        String resName = r?.name ?: "UNKNOWN"
        log.debug "Attempting to run action [$action] on [$resName]"
        ResourceApi resApi = api.resourceApi
        ControlApi cApi = api.controlApi

        long ctrlStart
        def resp = cApi.executeAction(r, action, args)
        if (resp) {
            ctrlStart = System.currentTimeMillis()
            log.debug "Submitted [$action] control action to [$resName]..."
        } else {
            log.error "[$action] failed for [$resName]:"
            log.error "\t${resp.error.reasonText}"
            return // from closure
        }

        long maxActionTime = System.getProperty("hq.control.maxTime","300000") as long
        boolean complete = false, failed = false
        long runTime = 0
        long startTime = System.currentTimeMillis()
        ControlHistory entry = null
        while (!complete && !failed && runTime<maxActionTime) {
            entry = getControlHistory(cApi.getHistory(r).controlHistory, action, ctrlStart)
            if (entry) {
                log.debug "Status of [$resName]:[$action] is [${entry.status}]"
                if (entry.status == "Completed") {
                    complete = true
                    continue
                } else if (entry.status == "Failed") {
                    failed = true
                    continue
                } else {
                    Thread.sleep(2000)
                }
            } else {
                log.warn "Unable to find recent control action history for [$action] on [$resName]"
            }
            runTime = System.currentTimeMillis() - startTime
        }
        if (complete) {
            log.info "Successfully processed [$action] on [$resName]"
            log.debug "\t$entry.message"
        } else {
            if (failed) {
                log.error "ERROR: [$action] on [$resName] failed"
                log.error "\t$entry.message"
            } else {
                log.error "[$action] on [$resName] timed out. Action may still complete at a later time."
            }
        }

        log.debug "runControlAction returning ${entry?.dump()}"
        return entry
    }

    /**
     * Finds the matching ControlHistory in the provided List for the given
     * action and (fuzzy) startTime.
     *
     * @param history The history to search through
     * @param action The name of the control action to search for
     * @param startTime The approximate start time of the control action
     *
     * @return the matching ControlHistory or null if it could not be found
     */
    ControlHistory getControlHistory(List<ControlHistory> history, String action, long startTime) {
        ControlHistory result = null
        long MAX_FUZZ = 5000L
        long fuzz = 0
        history.each { hist ->
            fuzz += 500
            if (hist.action == action) {
                if (Math.abs(startTime - hist.startTime) < MAX_FUZZ) {
                    result = hist
                }
            }
        }
        result
    }

    /**
     * Returns The list of resource instances of the given type
     *
     * @param type The name of the resource prototype
     *
     * @return The list of resource instances of the given type. This method
     *         returns an empty list if there are no instances or if the type
     *         does not exist.
     */
    List<Resource> getResources(String type) {
        List resources = []
        if (type) {
            try {
                def protoResp = api.resourceApi.getResourcePrototype(type)
                if (protoResp) {
                    def proto = protoResp.resourcePrototype
                    resources = api.resourceApi.getResources(proto, true, false).resource
                }
            } catch (Exception e) {
                log.warn "Unable to find resources of type [$type]", e
            }
        }

        return resources
    }

    def getSingletonResource(String resType) {
        List resources = getResources(resType)
        if (!resources) {
            log.warn "No resources of type [$resType] found"
            return null
        }
        def res = resources[0]
        if (resources.size() > 1) {
            log.warn "Multiple resources of type [$resType] found"
            return null
        }

        log.debug "Single resource of type [$resType]: ${res.dump()}"
        return res
    }

    boolean updateProperties(def res, Map newProps) {
        boolean propFound = false
        res.resourceProperty.each { resProp ->
            if (resProp.key in newProps.keySet()) {
                propFound = true
                resProp.value = newProps[resProp.key]
            }
        }
        if (propFound) {
            def status = api.resourceApi.updateResource(res)
            if (status) {
                log.debug "Successfully updated resource [$res.name] properties with: $newProps"
                return true
            } else {
                log.warn "Unable to update resource [$res.name]: ${status.error?.reasonText}"
            }
        } else {
            log.warn "Unable to update resource: The resource [$res.name] does not contain any properties in ${newProps?.keySet()}"
        }

        log.debug "The resource properties for [$res.name] were not updated with: $newProps"
        return false
    }

    // Finds a value in a list of objects that have getKey()/getValue()
    String getResourceProperty(Resource res, String key) {
        List list = res?.resourceProperty
        String result = null
        if (list && key) {
            list.each {
                if (it.key == key) {
                    result = it.value
                }
            }
        }

        return result
    }

    boolean updateConfig(def res, Map newConfig) {
        boolean configFound = false
        res.resourceConfig.each { resConfig ->
            if (resConfig.key in newConfig.keySet()) {
                configFound = true
                resConfig.value = newConfig[resConfig.key]
            }
        }
        if (configFound) {
            def status = api.resourceApi.updateResource(res)
            if (status) {
                log.debug "Successfully updated resource [$res.name] config with: $newConfig"
                return true
            } else {
                log.warn "Unable to update resource [$res.name]: ${status.error?.reasonText}"
            }
        } else {
            log.warn "Unable to update resource: The resource [$res.name] does not contain any config keys in ${newConfig?.keySet()}"
        }

        log.debug "The resource config for [$res.name] were not updated with: $newConfig"
        return false
    }

    // Finds a value in a list of objects that have getKey()/getValue()
    String getResourceConfig(Resource res, String key) {
        List list = res?.resourceConfig
        String result = null
        if (list && key) {
            list.each {
                if (it.key == key) {
                    result = it.value
                }
            }
        }

        return result
    }

    /**
     * Returns a List of Strings of the fully-qualified domain name (FQDN) of all
     * hosts running an agent connected to the SAS Environment Manager Server.
     * The FQDN is the unique identifier that can be used to find a platform resource
     * via the HQApi.
     *
     * @return a List of Strings of fully-qualified domain names
     */
    List getPlatforms() {
        api.agentApi.agents.agent.collect { it.address }
    }

    /**
     * INTERNAL METHOD
     * Obtains default connection information for SAS Environment Manager
     *
     * @return A map with the keys 'hostname', 'isSecure', 'port', 'user',
     *         'password', 'metadatahost', and 'metadataport' specifying
     *         connection information for SAS Environment Manager and Metadata
     */
    @Memoized
    Map getConnectionInfo() {
        Map result = [:]
        Properties agentProps = new Properties()
        Properties sasProps = new Properties()
        try {
            new File(getAgentDir(), "conf/agent.properties").withInputStream { stream ->
                agentProps.load(stream)
            }
            new File(getAgentDir(), "conf/sas.properties").withInputStream { stream ->
                sasProps.load(stream)
            }
        } catch (Exception e) {
            log.error("There was a problem obtaining SAS credentials", e)
            e.printStackTrace()
            result = ["hostname":"UNKNOWN", "isSecure":false, "port":-1, "user":"UNKNOWN", "password":"UNKNOWN", "metadatahost":"UNKNOWN", "metadataport":-1]
            return result
        }

        result["hostname"] = agentProps["agent.setup.camIP"]
        result["isSecure"] = "yes".equalsIgnoreCase(agentProps["agent.setup.camSecure"])
        result["user"] = sasProps["sas.meta.user"]
        if (result["isSecure"]) {
            result["port"] = agentProps["agent.setup.camSSLPort"] as Integer
        } else {
            result["port"] = agentProps["agent.setup.camPort"] as Integer
        }
        result["password"] = sasProps["sas.meta.pwd"]

        // Find a reachable Metadata Server - iterate over cluster members if necessary
        String metahost = null
        String metaport = null
        String[] metaHosts = sasProps["sas.meta.host"].split("\\s+")
        String[] metaPorts = sasProps["sas.meta.port"].split("\\s+")
        if (metaHosts.size() != metaPorts.size()) {
            log.error "The Metadata connection information is invalid"
            log.error "The number of hosts does not match the number of ports:"
            log.error "\t${metaHosts as List}"
            log.error "\t${metaPorts as List}"
        }
        for (int i=0; i<metaHosts.size(); i++) {
            String host = metaHosts[i]
            String port = metaPorts[i]
            // Test to see if the Metadata Server is responsive
            InetAddress address
            try {
                address = InetAddress.getByName(host)
                if (address.isReachable(10000)) {
                    // Success
                    log.debug "Found valid Metadata Server: $host:$port"
                    metahost = host
                    metaport = port
                } else {
                    log.warn "SAS Metadata Server [$host] failed to respond"
                }
            } catch (Exception e) {
                e.printStackTrace()
                log.warn "SAS Metadata Server [$host] failed getByName query"
            }
        }

        // Just use the first host if we couldn't find one
        // Returning nulls will probably cause more problems than it would try to solve
        result["metadatahost"] = metahost ?: metaHosts[0]
        result["metadataport"] = metaport ?: metaPorts[0]

        return result
    }
}
