/*****************************************************************************
 * Copyright  2015, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
 ****************************************************************************/

import java.net.InetAddress
import java.net.UnknownHostException

import groovy.transform.Memoized

import org.hyperic.sigar.Sigar
import org.hyperic.sigar.SigarException

import org.apache.commons.logging.Log

class NetworkUtil
{
    protected EMIScript emi
    protected Log log, msg, term
    protected sysProps = System.getProperties()
    protected env = System.getenv()
    protected Sigar sigar

    static final IPV4_REGEX = /^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/

    NetworkUtil(EMIScript baseScript) {
        if (!baseScript) {
            throw new IllegalArgumentException("Parent script instance required")
        }
        this.emi = baseScript
        this.log = emi.log
        this.msg = emi.msg
        this.term = emi.term
    }

    @Memoized
    public String getShortName() {
        String sName = fqdn
        if (isIPV4(sName)) {
            sName = emi.hq.sigar.netInfo.hostName
        }

        if (!isIPV4(sName)) {
            sName = sName?.split('\\.')[0]
        }

        return sName
    }

    @Memoized
    public String getFqdn() {
        String hostname = emi.hq.agentProperties['platform.fqdn']
        if (!hostname) {
            try {
                hostname = InetAddress.localHost.canonicalHostName
                // Sometimes, canonicalHostName can return the short name
                if (!hostname?.contains('.')) {
                    // Sigar does a better job of finding the FQDN
                    // Even if it can't, it falls back to the IP address
                    // See SigarPlatformDetector.java in the Hyperic source
                    // This is how the Agent determines the FQDN
                    hostname = emi.hq.sigar.FQDN
                }
            } catch(Exception e) {
                log.debug "Could not obtain FQDN the normal way. Using fallback", e
                try {
                    hostname = emi.hq.sigar.netInfo.hostName
                } catch (SigarException se) {
                    hostname = "localhost.unknown"
                    log.debug "Could not obtain FQDN at all. Using $hostname", se
                }
            }
        }

        log.debug "The fully-qualified domain name (fqdn) is [$hostname]"
        return hostname
    }

    @Memoized isIPV4(String ip) {
        return ip ==~ IPV4_REGEX
    }

    @Memoized
    public boolean isLocal(String hostname) {
        boolean isLocal = false

        if (hostname == "") {  // degenerate case, empty name
            return false
        }

        try {
            InetAddress.getAllByName(hostname).each { addr ->
                if (addr.hostAddress in localIPs) {
                    isLocal = true
                    return // from closure
                }
            }
        } catch (UnknownHostException hostEx) {
            log.warn("Unable to resolve host [" + hostname + "] ($hostEx)")
        }

        log.debug("isLocal [" + hostname + "] = " + isLocal)
        return isLocal
    }

    @Memoized
    // TODO: How does IPv6 play here?
    private List getLocalIPs() {
        List addrs = []

        // Add local IP overrides from agent properties (same as SAS Java plugins)
        String s = emi.hq.agentProperties["custom.local.ips"]
        if (s) {
            String[] overrideAddrs = s.split(",")
            for (String overrideAddr : overrideAddrs) {
                overrideAddr = overrideAddr.trim()
                if (overrideAddr.length() > 0) {
                    log.debug("getLocalIPs: Adding override IP: " + overrideAddr)
                    addrs << overrideAddr
                }
            }
        }

        // Get local IP addresses
        addrs << "127.0.0.1"
        try {
            String localHostname = InetAddress.getLocalHost().getCanonicalHostName()
            for (InetAddress localAddr : InetAddress.getAllByName(localHostname)) {
                addrs << localAddr.getHostAddress()
            }
            log.info("Local IPs for [" + localHostname + "] are " + addrs)
        } catch (UnknownHostException hostEx) {
            log.warn("Unable to determine local IP addresses ($hostEx).  Using default of 127.0.0.1")
        }

        return addrs
    }
}
