/*****************************************************************************
 * Copyright © 2015, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
 ****************************************************************************/

// NOTE: Requires Service Architecture runtime and will not function standalone

/* configureAPM.groovy - Traverses the levroot hierarchy, discovering paths
 * for supported SAS server types, adds -logconfigloc options for APM logging,
 * and sets required Metadata for ARM enablement in STP, PWS, and others. */

@BaseScript(EMIClientScript)
import groovy.transform.BaseScript
import org.apache.commons.cli.Option

import java.io.File
import java.text.SimpleDateFormat
import groovy.json.*
import groovy.io.FileType

// MetadataServer stuff
import com.sas.metadata.*
import com.sas.metadata.remote.*

String pgmName = "configureAPM"

debug = false
verbose = false
force = false
onEMIFrameworkHost = false

BEGIN_MARKER = "/* EMI_BEGIN */"
END_MARKER = "/* EMI_END */"

workspace = null
connection = null
reposFQID = null
shortReposId = null
objStore = null
_factory = null
errorCount = 0

def cli = new CliBuilder(
    usage: "${pgmName} [options] [--levroot <path>]",
    header: "Traverses the levroot hierarchy, discovering paths for supported SAS server types\n" +
            "and enables ARM setup and logging in various SAS EBI servers and Metadata.\n" +
            "Options:",
    footer: "Exit status:\n" +
            "$RC_SUCCESS  if OK\n" +
            "$RC_WARNING  if minor issues are encountered\n" +
            "$RC_GENERAL_ERROR  if an error is encountered\n" +
            "$RC_INVALID_USAGE for invalid usage"
)

cli.with {
    _(longOpt:'levroot', 'override default levroot path definition', args: 1, required: false)
    f(longOpt: 'force', 'Force the APM ARM config content to be removed and rewritten', required: false)
    v(longOpt:'verbose', 'Enable verbose output', required: false)
    d(longOpt:'debug', 'Enable debug output', required: false)
    h(longOpt:'help', 'Show usage information', required: false)
}

def opt = cli.parse(args)
if (!opt) {
    term.error "Invalid options"
    return exitPgm(RC_INVALID_USAGE)
}
if (opt.h) {
    cli.usage()
    return exitPgm(RC_SUCCESS)
}
if (opt.d) {
    debug = true
    setLogLevel LogLevel.DEBUG
    msg.debug "Debug Logging level set to DEBUG"
}
if (opt.v) {
    verbose = true
}
if (opt.f) {
    force = true
}
if (opt.levroot) {
	sysProps['sas.levroot'] = opt.levroot
	msg.debug "Level root override: [${opt.levroot}]"
}

msg.info "Service Architecture Client APM Configuration - ${new Date()}"

// Check to see if we're on the emi-framework machine
if (sysProps['sasev.dir']) {
	File emiframeworkdir = canonicalFile(sysProps['sasev.dir'], "emi-framework")
	if(emiframeworkdir.exists()) {
		onEMIFrameworkHost = true
		msg.debug "Running on the emi-framework host"
	} else {
		onEMIFrameworkHost = false
		msg.debug "Not running on the emi-framework host"
	}
} else {
	onEMIFrameworkHost = false
	msg.warn "Missing required property 'sasev.dir'. Cannot continue."
	return exitPgm(RC_GENERAL_ERROR)
}

emiConfigure(force)

msg.info "Service Architecture Client APM Configuration completed"
return RC_SUCCESS

//*******************
//*** Support Methods
//*******************
def FetchContextCollectionTable() {
	// first field is the general logging configuration type (AUDIT, PERF, GRID, ...)
	// second field is the SAS server name - this is the directory under Levroot/<context>

	ContextCollectionTable = [
//		['TYPE', 'ServerSubdir'],
		['AUDIT_META', 'MetadataServer'],
		['PERF', 'BatchServer'],
		['PERF', 'ConnectServer'],
		['GRID', 'GridServer'],  // renames existing logconfig files rather than new -logconfigloc
		['PERF_AUDIT_OLAP', '^OLAPServer\\d*'],
		['PERF', 'PooledWorkspaceServer'],
		['PERF', 'StoredProcessServer'],
		['PERF_WS', 'WorkspaceServer']  // requires different permissions on PerfLogs
		]
		return ContextCollectionTable
}

def FetchServerCollectionTable() {
	// first field is the general logging configuration type (AUDIT, PERF, GRID, ...)
	// second field is the SAS server flag file name - at this point, this is assumed to
	//   be one level below <levroot>. This could be expanded to recursive search into lower levels.
	//   For UNIX machines, this is "<server>.srv". For Windows machines, this is "<server>.bat".

	String flagext
	if(isWindows()) {
		flagext = ".bat"
		log.debug "isWindows: Flag file extension is '$flagext'"
	} else {
		flagext = ".srv"
		log.debug "isUnix: Flag file extension is '$flagext'"
	}

	ServerCollectionTable = [
//		['TYPE', 'ServerSRVfile'],
		['AUDIT_CONNECT', "ConnectSpawner${flagext}"],
		['AUDIT_OBJSPAWN', "ObjectSpawner${flagext}"]
		]
		return ServerCollectionTable
}

Properties propsFromStore(def txtStore) {
	def p = new Properties()
	p.load(new StringReader(txtStore))
	return p
}


int connect(host, port, user, password, repositoryStr) {
	if(verbose) {
		log.info "Connecting to metadata server:"
		log.info "    Host: $host"
		log.info "    Port: $port"
		log.info "    User: $user"
	}

	// Enable/Disable print of xml between client and server
	try {
		_factory = MdObjectFactory.getInstance()
		log.debug "factory is $_factory"
		_factory.setLoggingEnabled(false);
	} catch (Exception e) {
		term.error "A problem was encountered initializing the metadata server factory:\n ${e.message}"
		log.error "A problem was encountered initializing the metadata server factory\n", e
		return -1
	}

	// Connect to the metadata server
	try {
		workspace = MetadataWorkspace.getWorkspace();
		connection = workspace.makeOMRConnection(host, port, user, password);
		workspace.setCMRHandle(connection);
	} catch (Exception e) {
		term.error "A problem was encountered connecting to the metadata server:\n ${e.message}"
		log.error "A problem was encountered connecting to the metadata server\n", e
		return -2
	}

	if(verbose) msg.info "Connected to metadata server"

	// Get the repository information
	reposFQID = null;
	shortReposId = null;

	try {
		def repositories = MetadataUtil.getRepositories();
		repositories.each { repos ->
			if (repositoryStr.equals(repos.getName())) {
				reposFQID = repos.getFQID()
				shortReposId = reposFQID.substring(reposFQID.indexOf(".")+1)
				if(verbose) {
					 log.info "Found repository: $repositoryStr id: $reposFQID"
					 log.info "Short reposId: $shortReposId"
				}
			}
		}
	} catch (Exception e) {
		term.error "A problem was encountered fetching the metadata server repository information:\n ${e.message}"
		log.error "A problem was encountered fetching the metadata server repository information\n", e
		return -3
	}

	if (reposFQID == null || shortReposId == null) {
		log.error "Error finding repository named: $repositoryStr"
	}

	// Create the joma object store
	try {
		objStore = MdObjectFactory.createObjectStore();
	} catch (Exception e) {
		term.error "A problem was encountered creating the metadata server object store:\n ${e.message}"
		log.error "A problem was encountered creating the metadata server object store\n", e
		return -4
	}

	return 0  // all is well
}

boolean disconnect() {
	// Disconnect from the metadata server
	try {
		workspace.closeOMRConnection();
	} catch (Exception e) {
		term.error "A problem was encountered disconnecting from the metadata server:\n ${e.message}"
		log.error "A problem was encountered disconnecting from the metadata server\n", e
		return false
	}

	return true
}

def getMetadataObjects(type, search) {
	def xmlSelect = "<XMLSelect search=\"" + type + "[" + search + "]\" />"

	if(verbose) {
		log.info "Searching for metadata"
		log.info "    Type: $type"
		log.info "    Search: $xmlSelect"
	}

// working pattern = OMI_XMLSELECT | OMI_ALL | OMI_ALL_SIMPLE | OMI_INCLUDE_SUBTYPES | OMI_GET_METADATA
// OMI_ALL gets all properties
// OMI_ALL_SIMPLE gets all attributes
// OMI_INCLUDE_SUBTYPES gets all objects that are subtypes of the requested type also
// OMI_GET_METADATA iterates a GetMetadata call for all objects returned by GetMetadataObjects
// OMI_XML_SELECT uses the options parameter to filter/qualify the objects returned by the request

	def List objects = []
	try {
		objects = MetadataUtil.getMetadataObjectsSubset(objStore,
							reposFQID,
							type,
							MetadataUtil.OMI_XMLSELECT |
       							MetadataUtil.OMI_ALL |
							MetadataUtil.OMI_ALL_SIMPLE |
       							MetadataUtil.OMI_INCLUDE_SUBTYPES |
							MetadataUtil.OMI_GET_METADATA,
							xmlSelect)

	} catch (Exception e) {
		term.error "A problem was encountered fetching objects from the metadata server:\n ${e.message}"
		log.error "A problem was encountered fetching from the metadata server\n", e
		return null
	}


	def ocnt = objects.size()
	if(verbose) {
		log.info "Found metadata object count: $ocnt"

		if(debug) {
			objects.each { obj ->
				def oname = obj.getName()
				log.debug "    Object Name: $oname"
			}
		}
	}

	return objects;
}

def getMetadataObject(type, searchStr) {
	def retval = null;

	def objects = getMetadataObjects(type, searchStr);
	if (objects != null && objects.size() > 0) {
		retval = objects.get(0);
	}

	return retval;
}

boolean isConnected() {
	return (workspace != null);
}

int propertyUpdateCreate(def store, def obj, String matchname, String defaultval, String propname, boolean force) {
	log.debug "propertyUpdateCreate: object = $obj"
	log.debug "****  associations"

	int rc = -99  // arbitrary flag value, signals try-catch exception caught

	try {
		def Map lassoc = obj.getAssocs()
		def assoclist = lassoc["Properties"]
		log.debug "assoclist class=${assoclist.class} $assoclist"

		def propsparent = assoclist.getParent()
		log.debug "propsparent class=${propsparent.class}  $propsparent"

		boolean existingProp = false

		assoclist.each { myprop ->
			def Map myattr = myprop.getAttrs()
			log.debug "   assocattr:      $myprop, $myattr"

			if (myprop.name == matchname) {
				existingProp = true
				def curval = myprop.getDefaultValue()
				log.debug "myprop class=${myprop.class} $myprop"
				log.debug "myprop curval = $curval"

	// check to see if correct value already set
				if(curval == defaultval) {
					if(!force) {
						log.info "No Update: property value '$matchname' is correct"
						log.info "Existing value is '$curval'\n"
						rc = 0 // 0 == updated existing prop
						return // continue closure - matchname should be unique so rc will persist
					}
				}

	// Update the cached copy
				myprop.setDefaultValue(defaultval)
				myprop.setPropertyName(propname)
	// verify that the update was signaled (should always happen)
				if(myprop.objectModified()) {
					log.debug "Shows as modified"
					log.debug "issuing update"
	// write the cached update back to the server
					myprop.updateMetadataAll()

					if(force) {
						log.info "Force Update: property value '$matchname'"
						log.info "Previous value was '$curval'\n"
					} else {
						log.info "Update: property value '$matchname'"
						log.info "Previous value was '$curval'\n"
					}
					rc = 0 // 0 == updated existing prop
					return // continue closure - matchname should be unique instace so rc will persist
				}
				else {
	// should be a "can't happen" unless the server connection broke, etc. caught by try-catch
					log.error "Updated property shows as NOT modified in the object store"
					rc = -1  // n<0 == error, -1 == modification of existing obj failed
					return  // continue closure - matchname should be unique so rc will persist
				}
			}
		}

		if(existingProp)  // found something, return whether update succeeded or failed
			return rc

	// If we came this far, existing instance wasn't found; create a new property
		log.debug "Attempting to create PropertyType String"
		def mystrprop = _factory.createComplexMetadataObject(objStore,
					"String",
					MetadataObjects.PROPERTYTYPE,
					shortReposId)
					log.debug "Created object is class=${mystrprop.class} $mystrprop"
		if(mystrprop) {
			mystrprop.setSQLType(12)
		} else {
			log.error "Couldn't create a String PropertyType object"
			return -98  // exit method
		}

		log.debug "Attempting to create Property $matchname"
		def myprop = _factory.createComplexMetadataObject(objStore,
					matchname,
					MetadataObjects.PROPERTY,
					shortReposId)
		log.debug "Created object is class=${myprop.class} $myprop"
		if(myprop) {
			myprop.setDefaultValue(defaultval)
			myprop.setPropertyName(propname)
			myprop.setOwningType(mystrprop)
			myprop.setIsUpdateable(1 as int)
			myprop.setIsVisible(1 as int)
	// verify that the update was signaled (should always happen)
			if(myprop.objectModified()) {
				log.debug "Shows as modified"
				log.debug "issuing update"
	// write the cached update back to the server
				myprop.updateMetadataAll()
			}
			else {
	// should be a "can't happen" unless the server connection broke, etc. caught by try-catch
				log.error "Created property shows as NOT modified in the object store"
				return -2  // n<0 == error, -2 == modification of new obj failed, exit the method
			}
		} else {
			log.error "Couldn't create object for '$matchname'"
			return -3  // n<0 == error, -3 == couldn't create new object, exit the method
		}

	// add the new property object to the AssociationList for the parent object
		assoclist.addElement(myprop)
		if(obj.objectModified())  {
			log.debug "Main object shows as modified"
			log.debug "issuing update"
	// write the cached update back to the server
			obj.updateMetadataAll()
			log.info "Created: property value '$matchname'"
			rc = 1 // 1 == created and updated new prop
		} else {
	// should be a "can't happen" unless the server connection broke, etc. caught by try-catch
			log.error "Main object shows as NOT modified in the object store"
			rc = -2  // n<0 == error, -2 == modification of new obj failed
		}
	} catch (Exception e) {
		term.error "A problem was encountered creating/updating a property on the metadata server:\n ${e.message}"
		log.error "A problem was encountered creating/updating a property on the metadata server\n", e
		rc = -99
	}

	return rc
}


boolean emiConfigure(boolean force) {

// Get the patterns of files that we should collect
	def ContextCollectionTable = FetchContextCollectionTable()
	def ServerCollectionTable = FetchServerCollectionTable()

        LevRoots = []


// in the emi-client form of discovery, we know the local levelroot via a sysProp.
// Assemble an internal representation that's compatible with the legacy discovery codebase

	def levDir = sysProps['sas.levroot']
	def hostType = sysProps['os.name']
	def hostFQDN = InetAddress.getLocalHost().getCanonicalHostName().toLowerCase()

	log.debug "levdir is $levDir"
	log.debug "host type is $hostType"
	log.debug "FQDN is $hostFQDN"

	LevRoots << [hostFQDN, hostType, levDir]

// Generate a new APM configuration for the local machine elements
	if(LevRoots) {
		configureAPM(LevRoots, ServerCollectionTable, ContextCollectionTable, force)
	}

// Are we on the emi-framework host? If not, we're done with processing.

	if(!onEMIFrameworkHost) {
		msg.info "Not on the emi-framework host, skipping metadata update."
		return true
	}

// If we *are* on the emi-framework host, also configure necessary metadata for the Logical Pooled Workspace Servers,
// Logical Stored Process Servers, and Batch Servers.


// Establish connection to the Metadata Server to query for "levdirroot" objects

// The hq.connectionInfo comes in through our groovy master "run" script
	def myHost = hq.connectionInfo.metadatahost
	def myPort = hq.connectionInfo.metadataport
	def myUser = hq.connectionInfo.user
	def myPass = hq.connectionInfo.password

	if(!myHost || !myPort || !myUser || !myPass) {
		msg.error "The inherited Metadata Server connection info is invalid:"
		msg.error "host = '$myHost'"
		msg.error "port = '$myPort'"
		msg.error "user = '$myUser'"
		msg.error "pass = (deleted)"
		return false
	}

	connect(myHost,
		myPort,
		myUser,
		myPass,
		"Foundation")

	if(!isConnected()) {
		msg.error "Metadata Server Connection attempt failed."
		return false
	}

	try {
                def pwsobjs = getMetadataObjects(MetadataObjects.LOGICALSERVER,"@PublicType contains 'LogicalServer.PooledWorkspace'")

		def pwscnt = pwsobjs.size()
		log.info "****************************"
		(verbose? msg:log).info "Found $pwscnt 'LogicalServer.PooledWorkspace' objects in metadata"

                if(pwsobjs) {

			pwsobjs.each { thisobj ->
				int rc
				log.debug "object:  $thisobj"
				def Map lattrs = thisobj.getAttrs()
				if(debug)
					lattrs.each { log.debug "objattr:    $it" }
				log.debug "****  associations"
				def Map lassoc = thisobj.getAssocs()

				log.debug "Setting LPWS 'Initialization program' property"
				rc = propertyUpdateCreate(objStore,
					thisobj,
					"Initialization program",
					"SASEnvironment/SASMacro/apm_request_init.sas",
					"WorkspaceInit", force)
				log.debug "rc = $rc"

				log.debug "Setting LPWS 'Termination program' property"
				rc = propertyUpdateCreate(objStore,
					thisobj,
					"Termination program",
					"SASEnvironment/SASMacro/apm_request_term.sas",
					"WorkspaceTerm", force)
				log.debug "rc = $rc"
			}
		} // end if pwsobjs
	} catch (Exception e) {
		term.error "A problem was encountered creating/updating properties for LPWS:\n ${e.message}"
		log.error "A problem was encountered creating/updating properties for LPWS\n", e
		return false
	}


	try {
                def stpobjs = getMetadataObjects(MetadataObjects.LOGICALSERVER,"@PublicType contains 'LogicalServer.StoredProcess'")

		def stpcnt = stpobjs.size()
		log.info "****************************"
		(verbose? msg:log).info "Found $stpcnt 'LogicalServer.StoredProcess' objects in metadata"

                if(stpobjs) {

			stpobjs.each { thisobj ->
				int rc
				log.debug "object:  $thisobj"
				def Map lattrs = thisobj.getAttrs()
				if(debug)
					lattrs.each { log.debug "objattr:    $it" }
				log.debug "****  associations"
				def Map lassoc = thisobj.getAssocs()

				log.debug "Setting LSTP 'Initialization program' property"
				rc = propertyUpdateCreate(objStore,
					thisobj,
					"Initialization program",
					"SASEnvironment/SASMacro/apm_request_init.sas",
					"Request Init", force)
				log.debug "rc = $rc"

				log.debug "Setting LSTP 'Termination program' property"
				rc = propertyUpdateCreate(objStore,
					thisobj,
					"Termination program",
					"SASEnvironment/SASMacro/apm_request_term.sas",
					"Request Term", force)
				log.debug "rc = $rc"
			}
		} // end if stpobjs
	} catch (Exception e) {
		term.error "A problem was encountered creating/updating properties for LSTP:\n ${e.message}"
		log.error "A problem was encountered creating/updating properties for LSTP\n", e
		return false
	}


	try {
		// note: must use 'PublicType=' instead of 'contains' to avoid getting the Logical DSB server objects too
                def dsbobjs = getMetadataObjects(MetadataObjects.SERVERCOMPONENT,"@PublicType='Server.DataStepBatch'")

		def dsbcnt = dsbobjs.size()
		log.info "****************************"
		(verbose? msg:log).info "Found $dsbcnt 'Server.DataStepBatch' objects in metadata"

                if(dsbobjs) {

			dsbobjs.each { thisobj ->
				int rc
				log.debug "object:  $thisobj"
				def Map lattrs = thisobj.getAttrs()
				if(debug)
					lattrs.each { log.debug "objattr:    $it" }
				log.debug "****  associations"
				def Map lassoc = thisobj.getAssocs()

				log.debug "Setting LDSB 'BatchServer.DataStep.Property.LogExt.xmlKey.txt' property"
				rc = propertyUpdateCreate(objStore,
					thisobj,
					"BatchServer.DataStep.Property.LogExt.xmlKey.txt",
					"_",
					"LogExt", force)
				log.debug "rc = $rc"
			}
		} // end if dsbobjs
	} catch (Exception e) {
		term.error "A problem was encountered creating/updating properties for DSB:\n ${e.message}"
		log.error "A problem was encountered creating/updating properties for DSB\n", e
		return false
	}


	if(isConnected())
		disconnect()

	return true
}

boolean configureAPM(def lev, def ServerCollectionTable, def ContextCollectionTable, boolean force) {
	def contextDirs = []
	def serverPath = []
	def logs = []
	def dirCnt = 0	// total number of matching server directories discovered in this pass
	String serverType, serverContext, serverName, thisDirName

// Now discover configurable content in the levroot hierarchy(s)

	def is_first = true
	def found_it = false		// iterating through lists of files/directories
        def realContexts = []

	try {
		lev.each { thislev ->
			def levroot = new File(thislev[2])
			log.debug "iterating levroots: $levroot"

// create a list of contexts or subdirs at the levroot level
			levroot.eachDir() { contextDirs += it.getAbsoluteFile() }
		//	contextDirs = levroot.listFiles()

		/* yes, this could be reworked as an eachFileMatch closure filtering on type=dir */
			contextDirs.each { thisContextDir ->
				log.debug '*******************************'
				found_it = false
				thisDirName = thisContextDir.getName()

// Special-case: some server types don't live in a context subdir, so check for any of those first before descending
				ServerCollectionTable.each { logDef ->
					serverName = thisDirName
					def myCheckFile = new File(thisContextDir, logDef[1])

					if ( myCheckFile.exists()) {  // does it match the SAS server .srv name?
						log.debug "ServerCollection: serverPath=$thisContextDir serverName=$serverName"
						log.info "****************************"
						log.info "found: ${logDef[1]}\n        $thisContextDir"
						if(verbose) {
							term.info "****************************"
							term.info "found: ${logDef[1]}\n        $thisContextDir"
						}
						dirCnt++
						ProcessSubdir(is_first, thislev, thisContextDir, "", serverName, logDef)
						is_first = false
						found_it = true
						}
					} // each ServerCollectionTable

	// if found already in ServerCollection, bail out of the contextDirs closure early
				if (found_it) { return }

	// not found in ServerCollection, try in ContextCollection
				log.debug "Directory '$thisDirName' not found in ServerCollectionTable"
				log.debug "trying ContextCollectionTable"

				serverContext = thisDirName
				serverPath = thisContextDir.listFiles()
				log.debug "Potential serverContext is '$serverContext'"
				log.debug "Directory list is $serverPath"
				serverPath.each { thisPath ->
					log.debug "thisPath is $thisPath"
					if (thisPath.isDirectory()) {
						serverName = thisPath.getName() /* last part of the path name */
						ContextCollectionTable.each { logDef ->
// old version							if ( logDef[1] == serverName) {  // does it match the SAS server subdir name?
							if ( serverName ==~ logDef[1]) {  // does it match the SAS server subdir name?
								log.debug "ContextCollection: serverPath=$thisPath serverName=$serverName"
								log.info "****************************"
								log.info "found: ${logDef[1]}\n        $thisPath"
								if(verbose) {
									term.info "****************************"
									term.info "found: ${logDef[1]}\n        $thisPath"
								}
								dirCnt++
								ProcessSubdir(is_first, thislev, thisPath, serverContext, serverName, logDef)
								is_first = false
								if(logDef[0] != "AUDIT_META" ) {
									realContexts << [serverContext, thisContextDir]  // used later for SASEnvironment setup
								}
							}
						}
					}
				}  // serverPath.each
			}  //contextDirs.each
		} // lev.each

	} catch (Exception e) {
		term.error "A problem was encountered iterating through the levroot hierarchy:\n ${e.message}"
		log.error "A problem was encountered iterating through the levroot hierarchy\n", e
		return exitPgm(RC_GENERAL_ERROR)
	}
// For each unique appserver context, copy macros to SASEnvironment/SASMacros, and add ARM setup to *_autoexec_usermods.sas
	if (realContexts.size > 0) {
		realContexts.unique()
		log.info "****************************"
		msg.info "Found ${realContexts.size} unique app contexts."
		log.debug "realContexts = $realContexts"
		realContexts.each {thisContext ->
			ProcessContext(thisContext, force)
		}
        }

	log.info "****************************"
	msg.info "Discovered $dirCnt matching server directories."
}

boolean ProcessContext(def serverContext, boolean force) {
	log.info "****************************"
	log.info "Processing context '$serverContext'"
	if (verbose) term.info "Processing context '$serverContext'"

	String appusermodspath = "${serverContext[1]}${File.separator}appserver_autoexec_usermods.sas"
	updateUsermods(appusermodspath, "%apm_armsetup;", force)

	def sasEnvDir = new File(serverContext[1], "SASEnvironment").canonicalFile
	if (sasEnvDir.exists()) {
		def targetDir = new File(serverContext[1], "SASEnvironment/SASMacro").canonicalFile
		if (targetDir.canWrite()) {
			def srcpath = sysProps["emi.base.dir"]
			def srcDir = new File(srcpath, "content/AppContext/SASEnvironment/SASMacro").canonicalFile
			def myFiles = []

			try {
                        	srcDir.eachFile() { myFiles += it.canonicalFile }
				log.debug "ARM macros to copy = $myFiles"

				myFiles.each { thisFile ->
					def myName = thisFile.name
					def testFile = new File(targetDir, myName).canonicalFile
					if (testFile.exists() && !force) {
						log.info "No Update: file already exists: $testFile"
						return // note: effectively a 'continue' for the closure, not a method exit
					} else if(force) {
						log.info "Force Update: copying to: $testFile"
					} else {
						log.info "Update: copying to: $testFile"
					}
					// copy/overwrite the file
//msg.info "target is $testFile\nsrc is $thisFile\ncontent is ${thisFile.text}"
					testFile.write(thisFile.text)
				}

			} catch (Exception e) {
				term.error "A problem was encountered copying ARM macros to context's SASEnvironment:\n ${e.message}"
				log.error "A problem was encountered copying ARM macros to context's SASEnvironment\n", e
				return false
			}
		} else {
			msg.error "Cannot write to target SASMacros directory: $targetDir"
			return false
		}
		return true
	}
	return true  // If the SASEnvironment directory does not exist, it may be that this context has no executable code
}

void ProcessSubdir(Boolean is_first, def thislev, File thisPath, String serverContext, String serverName, def logDef) {
	log.debug "process: $thisPath, $serverContext, $serverName, $logDef"
	GString logLoc

	def usermodspath = "$thisPath${File.separator}sasv9_usermods.cfg"

	switch (logDef[0]) {	// logconfig type
		case 'AUDIT_META':
			log.info "AUDIT_META: $serverContext $serverName"
			if(verbose) term.info "AUDIT_META: $serverContext $serverName"
			createDir(thisPath, "AuditLogs", false)
			updateUsermods(usermodspath, force)
			break

		case 'AUDIT_CONNECT':
			log.info "AUDIT_CONNECT: $serverContext $serverName"
			if(verbose) term.info "AUDIT_CONNECT: $serverContext $serverName"
			createDir(thisPath, "AuditLogs", false)
// ConnectSpawner doesn't support SAS usermods, so directly replace logconfig.xml with APM version
			copySave(thisPath, "logconfig.apm.xml", "logconfig.xml", force, true)
			break

		case 'AUDIT_OBJSPAWN':
			log.info "AUDIT_OBJSPAWN: $serverContext $serverName"
			if(verbose) term.info "AUDIT_OBJSPAWN: $serverContext $serverName"
			createDir(thisPath, "AuditLogs", false)
// ObjectSpawner doesn't support SAS usermods, so directly replace logconfig.xml with APM version
			copySave(thisPath, "logconfig.apm.xml", "logconfig.xml", force, true)
			break

		case 'PERF_AUDIT_OLAP':
			log.info "PERF: $serverContext $serverName"
			if(verbose) term.info "PERF: $serverContext $serverName"
// OLAP has both Perf and Audit logs for APM
			createDir(thisPath, "PerfLogs", false)
			createDir(thisPath, "AuditLogs", false)
			updateUsermods(usermodspath, force)
			break

		case 'PERF':
			log.info "PERF: $serverContext $serverName"
			if(verbose) term.info "PERF: $serverContext $serverName"
			createDir(thisPath, "PerfLogs", false)
			updateUsermods(usermodspath, force)
			break

		case 'GRID':
			log.info "GRID: $serverContext $serverName"
			if(verbose) term.info "GRID: $serverContext $serverName"
// Grid is a special-case. It simply renames/copies existing ARM logconfig files into the appropriate names
// where they get automatically picked-up by the existing Grid run scripts.
			createDir(thisPath, "PerfLogs", false)
			copySave(thisPath, "logconfig.apm.batch.disable.xml", "logconfig.apm.batch.xml", force, false)
// don't create the flag file until the second copy is complete in this dir to avoid collision
			copySave(thisPath, "logconfig.apm.connect.disable.xml", "logconfig.apm.connect.xml", force, true)
			break

		case 'PERF_WS':
			log.info "PERF_WS: $serverContext $serverName"
			if(verbose) term.info "PERF_WS: $serverContext $serverName"
			createDir(thisPath, "PerfLogs", true)
			updateUsermods(usermodspath, force)
			break

		default:
			msg.error "ProcessSubdir: encountered unknown log configuration type: $logDef[0]"
			msg.error "serverName=$serverName, serverContext=$serverContext"
			msg.error "Terminating."
			exitPgm(RC_GENERAL_ERROR)
	}

}

boolean copySave(def thisPath, String src, String target, boolean force, boolean createFlag) {
	def srcFile = new File(thisPath, src).canonicalFile
	def targetFile = new File(thisPath, target).canonicalFile
	def savename = "${targetFile}_${_LOG_TIMESTAMP}"
        def saveFile = new File(savename)
	def flagFile = new File(thisPath, "emi_configured.txt").canonicalFile

// see if we've done this already - may override with -force option
	if(flagFile.exists()) {
		if (!force) {
			log.info "No Update: Already configured: $targetFile"
			return true
		} else {
			log.info "Force Update: $targetFile"
		}
	}

// save the current config file with an appended datetime stamp
	if(targetFile.exists()) {
		log.info "Copying original $targetFile to $savename"
		if(verbose)
			term.info "Copying $targetFile to $savename"

		try {
			def myText = targetFile.text	// in two steps just to aid future debugging
			saveFile.write(myText)
		} catch (Exception e) {
			term.error "A problem was encountered copying file: $targetFile\n ${e.message}"
			log.error "copySave: A problem was encountered copying file: $targetFile\n", e
			return false
		}
	}

// copy the current src file to the target file name
	log.info "Copying $srcFile to $targetFile"
	if(verbose)
		term.info "Copying $srcFile to $targetFile"
// Due to some complexities of UGIP, the src file may not exist in some special configurations.
// In that case, juse issue a warning, but continue on. This effectively yields a degraded APM such
// that the APM logconfig is not enabled, so no ARM logs will be collected, but the existing default
// logconfig.xml should still be in-place and work.
	if (srcFile.exists()) {
		try {
			def myText = srcFile.text	// in two steps just to aid future debugging
			targetFile.write(myText)
		} catch (Exception e) {
			term.error "A problem was encountered copying file: $srcFile\n ${e.message}"
			log.error "copySave: A problem was encountered copying file: $srcFile\n", e
			return false
		}
	} else {
		msg.warn "***** Attention *****"
		msg.warn "Expected APM logconfig file does not exist: $srcFile"
		msg.warn "configureAPM will leave the existing logconfig file in-place."
		msg.warn "As a result, no ARM logs will be generated for this server instance."
		msg.warn " "
		// proceed anyway so that the flag file gets created, allowing us to NO-OP this
	}

// create a flag file to show that we have configured APM already
	if(createFlag) {
		log.info "Creating flag file: $flagFile"
		if(verbose)
			term.info "Creating flag file: $flagFile"
		try {
			flagFile << "Flag file used to indicate that EV/APM configuration has been performed here.\n" +
					"DO NOT REMOVE FILE.\n"
		} catch (Exception e) {
			term.error "A problem was encountered creating flag file: $flagFile\n ${e.message}"
			log.error "copySave: A problem was encountered creating flag file: $flagFile\n", e
			return false
		}
	}
	return true
}

boolean createDir(def path, String subdir, boolean everyone_access) {
	def myDir = new File(path, subdir)
	def myDirName = myDir.canonicalPath
	def owner_only = !everyone_access	// logic is inverted on setWritable, etc.

	log.info "Creating ${myDir.canonicalPath}"

	if (myDir.exists()) {
		log.info "$myDir already exists."
	}
	else {
		try {
			myDir.mkdir()
		}
		catch (Exception e) {
			term.error "A problem was encountered creating ${myDirName}: ${e.message}"
			log.error "A problem was encountered creating ${myDirName}:", e
			return false
		}

	}


	try {  // set directory permissions
		if(isWindows()) {
			log.debug "Setting directory permissions (Windows) for: $myDirName"
			myDir.setReadable(true, owner_only)
			myDir.setWritable(true, owner_only)
			myDir.setExecutable(true, owner_only)
		} else {
			def access_mode = '0775'	// default is owner:group rwx
			if(everyone_access)
				access_mode = '0777'	// everyone rwx

			log.debug "Setting directory permissions (UNIX $access_mode) for: $myDirName"
			def myCmd = "chmod $access_mode $myDirName"
			rc = exec(myCmd)
			if(rc) {
				msg.error "There was a problem setting access permissions for $myDirName"
				msg.error "rc = $rc"
				return false
			}
		}

	}
	catch (Exception e) {
		term.error "A problem was encountered changing directory permissions for ${myDirName}: ${e.message}"
		log.error "A problem was encountered changing directory permissions for ${myDirName}:", e
		return false
	}

	return true
}



int updateUsermods(String usermodspath, boolean force) {
// in this form, the method derives the -logconfigloc path from the 'usermodspath' string

// this is somewhat overkill just to create some path strings, but Groovy handles the OS-specifics for us,
// so not reinventing the wheel.
	def myFile = new File(usermodspath)
	def myCfile = myFile.canonicalFile
	def logconfiglocpath = "-logconfigloc ${myCfile.parent}${File.separator}logconfig.apm.xml"

	msg.debug "updateUsermods(p,f):"
	msg.debug "   myFile=${myFile.name}"
	msg.debug "   myCfile=${myCfile}"
	msg.debug "   logconfiglocpath=$logconfiglocpath"

	return updateUsermods(usermodspath, logconfiglocpath, force)
}


int updateUsermods(String usermodspath, String addedText, boolean force) {
// in this form, the method uses the supplied fully-qualified paths to update sasv9_usermods.cfg

	// see if the EMI marker is already there
	def mymods = new File(usermodspath).canonicalFile
	def fileContent = []
	def foundBegin = false
	def foundEnd = false
	try {
		if (mymods.exists()) {
			log.debug "updateUsermods: Found $mymods"
// read the complete file into a collection, flagging if we see our "EMI_BEGIN" marker
			mymods.eachLine { line ->
				if (line.startsWith(BEGIN_MARKER))
					foundBegin = true
				else if (line.startsWith(END_MARKER))
					foundEnd = true
				fileContent << line
			}
			log.debug "foundBegin=$foundBegin"
			log.debug "foundEnd=$foundEnd"
			// log.debug "text is:\n$fileContent"
		} else {
			msg.error "Couldn't find expected usermods file at '$mymods'"
			return 2
		}
	} catch (Exception e) {
                term.error "A problem was encountered reading the usermods file:\n ${e.message}"
                log.error "A problem was encountered reading the usermods file\n", e
                return exitPgm(RC_GENERAL_ERROR)
	}

// Sanity check - should have neither or both Begin and End markers
	if(foundBegin || foundEnd) {
		if (!(foundBegin && foundEnd)) {
			msg.error "updateUserMods: didn't find matching Begin/End tags for EMI content: Begin=$foundBegin, End=$foundEnd"
			msg.info "Skipping $mymods"
			return 3
		}
	}  // end sanity check

// We have the file content buffered. Based on "force", decide whether to append, update, or NO-OP

	if (!foundBegin) {
		appendUsermods(mymods, fileContent, addedText)
		log.info "Append: EMI content added to $mymods"
		if(verbose)
			term.info "Append: EMI content added to $mymods"
	} else if (foundBegin && force) {
		def newContent = stripUsermods(fileContent)
		if(newContent) {
			msg.debug "stripped text is\n${newContent.join("\n")}\n"
			appendUsermods(mymods, newContent, addedText)
			log.info "Force Append: EMI content stripped and re-added to $mymods"
			if(verbose)
				term.info "Force Append: EMI content stripped and re-added to $mymods"
		} else {
			log.error "Force Append: couldn't remove existing EMI content in $mymods"
			log.error "Check that correct code blocking exists:"
			log.error "$BEGIN_MARKER"
			log.error "$END_MARKER"
			log.info "Skipping $mymods"
			return 3
		}

	} else {
		log.info "No Update: EMI content already in $mymods"
		if(verbose)
			term.info "No Update: EMI content already in $mymods"
		}

	return 0
}

// appendUsermods - add EMI content to end of sasv9_usermods.cfg file
//
// This code attempts to be safe by copying the existing content to a new temp file, appending EMI content to that, renaming
// the existing file with an appended datetime string, and renaming the temp file to the original name.
//
void appendUsermods(File myFile, def fileContent, String addedText) {
	log.debug "appendUsermods:"
	log.debug "myFile is $myFile"

	def curpath = myFile.parent
	def savename = "${myFile}_${_LOG_TIMESTAMP}"
	def tempFile = new File(curpath, "usermods.tmp")

// this is the code block appended to usermods to enable APM ARM logging
	def newcontent =
		"$BEGIN_MARKER\n" +
		"/* Added by configureAPM at ${_LOG_TIMESTAMP}, DO NOT EDIT THESE LINES OR EMI TAGS */\n" +
		"$addedText\n" +
		"$END_MARKER\n"

	try {
		tempFile << fileContent.join("\n") + "\n"  // final newline
		tempFile << newcontent
	} catch (Exception e) {
                term.error "appendUsermods: A problem was encountered writing the usermods.tmp file: $tempFile\n ${e.message}"
                log.error "appendUsermods: A problem was encountered writing the usermods.tmp file: $tempFile\n", e
                exitPgm(RC_GENERAL_ERROR)
        }
// now save the current usermods file with an appended datetime stamp
	try {
		myFile.renameTo( new File(savename))
        } catch (Exception e) {
                term.error "appendUsermods: A problem was encountered renaming file: $myFile\n ${e.message}"
                log.error "appendUsermods: A problem was encountered renaming file: $myFile\n", e
                exitPgm(RC_GENERAL_ERROR)
        }

// now set the revised usermods file to the original name - note "myFile" is immutaable, so the above rename did
// not change the original File name or definition, thus we can use that File object for our target
	try {
		tempFile.renameTo( myFile)
        } catch (Exception e) {
                term.error "appendUsermods: A problem was encountered renaming file: $myFile\n ${e.message}"
                log.error "appendUsermods: A problem was encountered renaming file: $myFile\n", e
                exitPgm(RC_GENERAL_ERROR)
        }
	log.info "Saved existing usermods file to $savename"
	if (verbose)
		term.info "Saved existing usermods file to $savename"
}

String [] stripUsermods(def fileContent) {
	def foundBegin = false
	def foundEnd = false
	def newContent = []

	fileContent.each {
		if(it.startsWith(BEGIN_MARKER)) {
			foundBegin = true
		} else if(it.startsWith(END_MARKER)) {
			foundEnd = true
		} else if(foundBegin && !foundEnd) {
		// ignore content while inside EMI code block
		} else {
			newContent << it
		}
	}
	log.debug "Stripped text:\n${newContent.join("\n")}\n"
	if(foundBegin && foundEnd) {	// found a correct bracket of EMI code
		log.debug "stripUserMods: Successfully stripped content"
		return newContent
	} else {
		msg.error "stripUserMods: didn't find matching Begin/End tags for EMI content: Begin=$foundBegin, End=$foundEnd"
		return null
	}
}

void ListProps(String what, Properties theseProps)
{
	PrintWriter writer = new PrintWriter(System.out)
	println(what)
	theseProps.list(writer)
	writer.flush()
}

def exitPgm(int rc) {
    msg.debug "configureAPM is exiting with return code $rc"
    System.exit rc
}
