/*****************************************************************************
 * Copyright © 2015, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.
 ****************************************************************************/

// NOTE: Requires Service Architecture runtime and will not function standalone

@BaseScript(EMIClientScript)
import groovy.transform.BaseScript
import org.apache.commons.cli.Option

import java.text.SimpleDateFormat
import groovy.json.*
import groovy.io.FileType

String pgmName = "logDiscovery"

boolean debug = false
apmInitOpt = false
configureAPM = true
force = false
fs = File.separator

def cli = new CliBuilder(
    usage: "$pgmName [-fhvd] [--levroot <path>] [--noconfigure] [--apmInit]",
    header: "Traverses the config hierarchy, discovering log paths for supported server types\n" +
            "Options:",
    footer: "Exit status:\n" +
            "$RC_SUCCESS  if OK\n" +
            "$RC_WARNING  if minor issues are encountered\n" +
            "$RC_GENERAL_ERROR  if an error is encountered\n" +
            "$RC_INVALID_USAGE for invalid usage"
)

cli.with {
    _(longOpt:'levroot', 'Override default level root path', args: 1, required: false)
    _(longOpt:'noconfigure', 'Disables invoking the ConfigureAPM utility as a last step', required: false)
    _(longOpt:'apmInit', 'Perform local APM initialization', required: false)
    f(longOpt:'force', 'Force the APM config content to be removed and rewritten', required: false)
    v(longOpt:'verbose', 'Enable verbose output', required: false)
    d(longOpt:'debug', 'Enable debug output', required: false)
    h(longOpt:'help', 'Show usage information', required: false)
}

def opt = cli.parse(args)
if (!opt) {
    term.error "Invalid options"
    return exitPgm(RC_INVALID_USAGE)
}
if (opt.h) {
    cli.usage()
    return exitPgm(RC_SUCCESS)
}
if (opt.d) {
    debug = true
    setLogLevel LogLevel.DEBUG
    msg.debug "Debug Logging level set to DEBUG"
}
if (opt.v) {
    verbose = true
}
if (opt.f) {
    force = true
}
if (opt.apmInit) {
    apmInitOpt = true
    log.info "*** Invoked with --apmInit, overidding state of apm.initialized"
}
if (opt.levroot) {
    sysProps['sas.levroot'] = opt.levroot
    msg.warn "Overriding default levroot definition with ${opt.levroot}"
}
if (opt.noconfigure) {
    configureAPM = false
}

msg.info "Service Architecture Log Discovery - ${new Date()}"

boolean apmInitialized = false

// check to see if the apm.initialized resource is set, if not found, tell the user and exit.
if (verbose) { msg.info "Verifying that APM is initialized..." }
log.debug "Fetching 'SAS Environment Manager Data Mart 9.4 APM ETL Processing' resource"
def apm = hq.getSingletonResource("SAS Environment Manager Data Mart 9.4 APM ETL Processing")
if (apm) {
    apmInitialized = (hq.getResourceProperty(apm, 'apm.initialized') ==~ /Yes/)
    (verbose? msg:log).info "APM Initialized = $apmInitialized"
} else {
    msg.error "Unable to find APM ETL Processing resource"
    msg.error "APM configuration and discovery will not be run."
    return exitPgm(RC_GENERAL_ERROR)
}

// Run if initialized, or force
if (apmInitialized || apmInitOpt) {
    if (configureAPM) {
        msg.info "Configuring APM..."
        confArgs = []
        if (verbose) confArgs << '-v'
        if (force) confArgs << '-f'
        if (debug) confArgs << '-d'
        if (opt.levroot) {
            confArgs << '--levroot'
            confArgs << opt.levroot
        }
        (verbose? msg:log).info "Calling configureAPM with options $confArgs"

        int rc = runGroovy(canonicalPath(scriptDir, 'configureAPM.groovy'), confArgs)
        if (rc != 0) {
            msg.error "There was a problem running configureAPM.groovy [$rc]). Check ${logFile.name}."
            return exitPgm(RC_GENERAL_ERROR)
        }
    } else {
        (verbose? msg:log).info "Skipping APM configuration"
    }

    msg.info "Running discovery..."
    int rc = emiDiscover()
    if (rc != 0) {
        msg.error "There was a problem while discovering log content (rc = $rc)"
        return exitPgm(RC_GENERAL_ERROR)
    }

    (verbose? msg:log).info "[log_locations.json] written to Conf"
    msg.info "Service Architecture Log Discovery completed successfully"
} else {
    msg.info "APM is not initialized. Skipping log discovery."
}

return exitPgm(RC_SUCCESS)

//*******************
//*** Support Methods
//*******************
def fetchCollectionTable() {
    return [
        // [type:'context', prefixPath:'%l%', serverPart:'RandomSubdir', logPart:'Logs', filePattern:'BaseLogPattern', operation:'Operation', namePattern:'NamePattern', descr:'Generic log descr'],
        [type:'context',  prefixPath:'%l%', serverPart:'MetadataServer', logPart:'AuditLogs', filePattern:'Access_SASMeta_MetadataServer.*', operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Metadata Server (Access)',   descr:'SAS Metadata Server Access Control activity'],
        [type:'context',  prefixPath:'%l%', serverPart:'MetadataServer', logPart:'AuditLogs', filePattern:'Audit_SASMeta_MetadataServer.*',  operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Metadata Server (Audit)',    descr:'SAS Metadata Server Audit activity'],
        [type:'context',  prefixPath:'%l%', serverPart:'BatchServer',    logPart:'PerfLogs',  filePattern:'arm4_BatchServer.*',              operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical SAS DATA Step Batch Server', descr:'SAS Batch Server ARM logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'ConnectServer',  logPart:'PerfLogs',  filePattern:'arm4_ConnectServer.*',            operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Connect Server',             descr:'SAS Connect Server ARM logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'ConnectServer',  logPart:'Logs',      filePattern:'Audit.Library_server.*',          operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Connect Server (DSA)',             descr:'SAS Connect Server DSA logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'ConnectServer',  logPart:'Logs',      filePattern:'Audit.MetaboundLib.*',            operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Connect Server (MBL)',             descr:'SAS Connect Server MBL logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'GridServer',     logPart:'PerfLogs',  filePattern:'arm4_GridServer.*',               operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Grid Server',                descr:'SAS Grid Server ARM logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'GridServer',     logPart:'Logs',      filePattern:'Audit.Library.*',                 operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Grid Server (DSA)',                descr:'SAS Grid Server DSA logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'GridServer',     logPart:'Logs',      filePattern:'Audit.MetaboundLib.*',            operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Grid Server (MBL)',                descr:'SAS Grid Server MBL logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'^OLAPServer\\d*',     logPart:'PerfLogs',  filePattern:'arm4_OLAPServer.*',               operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical OLAP Server',                descr:'SAS OLAP Server ARM logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'^OLAPServer\\d*',     logPart:'AuditLogs', filePattern:'Audit_OLAPServer.*',              operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical OLAP Server (Audit)',        descr:'SAS OLAP Server Audit logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'^OLAPServer\\d*',     logPart:'Logs',      filePattern:'Audit.Library_server.*',          operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical OLAP Server (DSA)',        descr:'SAS OLAP Server DSA logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'^OLAPServer\\d*',     logPart:'Logs',      filePattern:'Audit.MetaboundLib.*',            operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical OLAP Server (MBL)',        descr:'SAS OLAP Server MBL logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'PooledWorkspaceServer', logPart:'PerfLogs', filePattern:'arm4_PooledWSServer.*',  operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Pooled Workspace Server', descr:'SAS Pooled Workspace Server ARM logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'PooledWorkspaceServer', logPart:'Logs',     filePattern:'Audit.Library_server.*', operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Pooled Workspace Server (DSA)', descr:'SAS Pooled Workspace Server DSA logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'PooledWorkspaceServer', logPart:'Logs',     filePattern:'Audit.MetaboundLib.*',   operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Pooled Workspace Server (MBL)', descr:'SAS Pooled Workspace Server MBL logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'StoredProcessServer',   logPart:'PerfLogs', filePattern:'arm4_STPServer.*',       operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Stored Process Server',   descr:'SAS Stored Process Server ARM logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'StoredProcessServer',   logPart:'Logs',     filePattern:'Audit.Library_server.*', operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Stored Process Server (DSA)',   descr:'SAS Stored Process Server DSA logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'StoredProcessServer',   logPart:'Logs',     filePattern:'Audit.MetaboundLib.*',   operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Stored Process Server (MBL)',   descr:'SAS Stored Process Server MBL logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'WorkspaceServer',       logPart:'PerfLogs', filePattern:'arm4_(.+)',              operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Workspace Server',        descr:'SAS Workspace Server ARM logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'WorkspaceServer',       logPart:'Logs',     filePattern:'Audit.Library_server.*', operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Workspace Server (DSA)',        descr:'SAS Workspace Server DSA logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'WorkspaceServer',       logPart:'Logs',     filePattern:'Audit.MetaboundLib.*',   operation:'Copy', purgeAfter: "30", namePattern:'%c% - Logical Workspace Server (MBL)',        descr:'SAS Workspace Server MBL logging'],
        [type:'context',  prefixPath:'%l%', serverPart:'SASWebReportStudio',    logPart:'Logs',     filePattern:'WRS_KEYACTIONS.*', operation:'Move_closed', namePattern:'%c% - Logical Metadata Server', descr:'SAS Web Report Studio Key Actions logging'],
        [type:'absolute', prefixPath:'%l%', contextForce: 'nocontext', serverPart:'ConnectSpawner', logPart:'Logs',      filePattern:'ConnectSpawner_.*',     operation:'Copy', purgeAfter: "30", namePattern:'Logical Connect Spawner',          descr:'SAS Connnect Spawner activity'],
        [type:'dotsrv',   prefixPath:'%l%', contextForce: 'nocontext', serverPart:'ObjectSpawner',  logPart:'AuditLogs', filePattern:'Audit_ObjectSpawner.*', operation:'Copy', purgeAfter: "30", namePattern:'%n% - %h% Object Spawner (Audit)', descr:'SAS Object Spawner Audit activity'],
        [type:'absolute', prefixPath:'%l%/Web/SASEnvironmentManager/emi-client', contextForce:'kits', serverPart:'KitLogs', logPart:'SASJobs', filePattern:'.+.log', operation:'Move_all', namePattern:'EMI Kit SASJobs - %h% (Audit)', descr:'Generic SAS User jobs'],
        [type:'absolute', prefixPath:'%l%/Web', contextForce:'kits', serverPart:'WebServer', logPart:'logs', filePattern:'access.*', operation:'Move_closed', namePattern:'SAS Web Server Access', descr:'SAS Web Server access'],
        [type:'wildcard', prefixPath:'%l%/Web/WebAppServer', filePattern:'server.lo.*', logPart:'logs', logDest: 'kits/WebAppServer/%w%', operation:'Move_closed', namePattern:'SAS Web App Server activity', descr:'SAS Web App Server activity'],
        [type:'wildcard', prefixPath:'%l%/Web/Logs', filePattern:'.*VisualAnalytics.*', logDest: 'kits/VALASR/%w%', operation:'Move_closed', namePattern:'SAS Visual Analytics', descr:'SAS Visual Analytics activity']
    ]
}

Properties propsFromStore(def txtStore) {
    def p = new Properties()
    p.load(new StringReader(txtStore))
    return p
}

int emiDiscover() {
    // Get the patterns of files that we should collect
    def CollectionTable = fetchCollectionTable()
    def ContextCollectionTable = []
    def AbsoluteCollectionTable = []
    def WildcardCollectionTable = []
    def DotsrvCollectionTable = []
    LevRoots = []

    // populate the Context and Server tables with their respective elements of the Collection table
    CollectionTable.each { thisentry ->
        switch (thisentry.type) {
            case 'context':
                ContextCollectionTable << thisentry
                break
            case 'absolute':
                AbsoluteCollectionTable << thisentry
                break
            case 'wildcard':
                WildcardCollectionTable << thisentry
                break
            case 'dotsrv':
                DotsrvCollectionTable << thisentry
                break
            default:
                msg.warn "Unknown log collection type: ${thisentry.type}, ignored..."
                break
        }
    }

    // Assemble an internal representation that's compatible with the legacy discovery codebase
    def levDir = sysProps['sas.levroot']
    def hostType = sysProps['os.name']
    def hostFQDN = InetAddress.getLocalHost().getCanonicalHostName().toLowerCase()

    log.debug "levdir is $levDir"
    log.debug "host type is $hostType"
    log.debug "FQDN is $hostFQDN"

    LevRoots << [hostFQDN:hostFQDN, hostType:hostType, levDir:levDir]

    // Output the levroot definitions as a JSON file
    if (LevRoots) {
        createJSONfile(LevRoots, ContextCollectionTable, AbsoluteCollectionTable,
                       WildcardCollectionTable, DotsrvCollectionTable)
    }

    return 0
}

int createJSONfile(def lev, def ContextCollectionTable, def AbsoluteCollectionTable,
                   def WildcardCollectionTable, def DotsrvCollectionTable) {
    def contextDirs = []
    def serverPath = []
    def logs = []
    def dirCnt = 0 // total number of matching log directories discovered
    String serverType, serverContext, serverName

    // Some setup - make sure the emi-client/Conf dir exists
    def targetDir = canonicalPath(emiBaseDir, "Conf")

    try {
        def myDir = new File(targetDir)
        if (!myDir.exists()) {
            msg.error "Expected configuration directory does not exist: $targetDir"
            return exitPgm(RC_GENERAL_ERROR)
        }

        if (!myDir.canWrite()) {
            def myFpath = myDir.getCanonicalPath()
            msg.error "Cannot write to directory at $targetDir"
            return exitPgm(RC_GENERAL_ERROR)
        }
    } catch (Exception e) {
        term.error "A problem was encountered checking the existence of the Conf dir: $e.message"
        log.error "A problem was encountered checking the existence of the Conf dir", e
        return exitPgm(RC_GENERAL_ERROR)
    }

    // Now create and write the JSON file describing the levdirroot(s)
    def myFile
    try {
        myFile = new File(targetDir, "log_locations.json")
        if (myFile.exists()) {
            def myFpath = myFile.getCanonicalPath()
            log.info "Configuration file already exists at $myFpath, overwriting..."
        }

        // Write the opening bracket for the JSON content
        myFile.text = "[\n"
    } catch (Exception e) {
        term.error "A problem was encountered creating the definitions JSON file: $e.message"
        log.error "A problem was encountered creating the definitions JSON file", e
        return exitPgm(RC_GENERAL_ERROR)
    }

    def is_first = true // Have not written to file yet

    try {
        lev.each { thislev ->
            def levroot = new File(thislev.levDir)
            log.debug "iterating levroots: $levroot"

            // Special-case: some server types either don't live in a context
            // per se, or specify an absolute path to their content
            AbsoluteCollectionTable.each { logDef ->
                log.debug '*******************************'
                def found_it = false
                def absDir = logDef.prefixPath.replace('%l%', thislev.levDir) + fs +
                             logDef.serverPart + fs + logDef.logPart
                log.debug "AbsoluteCollection: absDir = '$absDir'"
                def absFile = new File(absDir).canonicalFile

                if (absFile.exists()) {
                    log.debug "AbsoluteCollection(found): serverPath=$absDir"
                    dirCnt++
                    ProcessLogs(myFile, is_first, thislev, absFile, "", "", logDef)
                    is_first = false
                    found_it = true
                }
            } // each AbsoluteCollectionTable

            // Special-case: some server types have variable names which we
            // need to wildcard to, similar to a context but without the
            // <levroot>/<context>/<servertype> pattern
            WildcardCollectionTable.each { logDef ->
                log.debug '*******************************'
                def wildDirPath = logDef.prefixPath.replace('%l%', thislev.levDir)
                log.debug "wildDirPath=$wildDirPath"
                def wildDirFile = new File(wildDirPath).canonicalFile
                log.debug "wildDirFile=$wildDirFile"
                def wildFiles = wildDirFile.listFiles()
                log.debug "wildFiles=$wildFiles"
                def found_it = false

                // Starting at the prefixPath, collect all file/directories at that level (ie - <prefixPath>/*)
                // For each element of "*", see if it contains logs that match optional logPart + the pattern "filePattern"
                // If there is a match, then add this log definition (once) to the log_locations.json file
                // If none match, continue
                wildFiles.each { thisWild ->
                    def matchWildFile = canonicalFile(thisWild)
                    if (logDef.logPart) {
                        matchWildFile = canonicalFile(matchWildFile, logDef.logPart)
                    }
		    log.debug "matchWildFile=$matchWildFile"

                    if (!found_it && matchWildFile.isDirectory()) {
                        matchWildFile.eachFileMatch(FileType.FILES, ~/${logDef.filePattern}/) {
                            found_it = true
                        }
			if(found_it) {
			    dirCnt++
			    ProcessLogs(myFile, is_first, thislev, thisWild, "", "", logDef)
			    is_first = false
			}
		    }
                }
            } // each WildcardCollectionTable

            // Special-case: some server types don't live in a context per se,
            // and may have multiple arbitrary-named subdirs
            // We find these by searching for associated <server>.srv files
            // (normally used by Lev1/generate_boot_scripts.sh)
            // For Windows machines, search for the associated script control
            // file instead <server.bat>.
            // At present, the search is limited to subdirs at the <levroot> level.
            // This could be expanded to a more-general case using a recursive
            // search like generate_boot_scripts.sh.
            String flagext = isWindows()? ".bat" : ".srv"
            log.debug "Flag file extension is '$flagext'"

            DotsrvCollectionTable.each { logDef ->
                log.debug '*******************************'

                // Create a list of contexts or subdirs at the levroot level
                contextDirs = [] // reset the list
                levroot.eachDir() { contextDirs += canonicalFile(it) }

                 // Yes, this could be reworked as an eachFileMatch closure filtering on type=dir
                 contextDirs.each { thisDir ->
                    log.debug '*******************************'

                    def found_it = false
                    def srvDir = logDef.prefixPath.replace('%l%', thislev.levDir) + fs + thisDir.name
                    log.debug "DotsrvCollection: srvDir = '$srvDir'"
                    def srvFile = new File(srvDir, "${logDef.serverPart}${flagext}").canonicalFile

                    if (srvFile.exists()) {
                        def myDir = new File(srvDir).canonicalFile
                        log.debug "DotsrvCollection(found): srvDir=$srvDir"
                        log.debug "DotsrvCollection(found): myDir=$myDir"
                        dirCnt++
                        ProcessLogs(myFile, is_first, thislev, myDir, "", "", logDef)
                        is_first = false
                        found_it = true
                    }
                } // each contextDirs
            } // each DotsrvCollectionTable

            // create a list of contexts or subdirs at the levroot level
            contextDirs = []    // reset the list
            levroot.eachDir() { contextDirs += it.getCanonicalFile() }
            // contextDirs = levroot.listFiles()

            /* yes, this could be reworked as an eachFileMatch closure filtering on type=dir */
            contextDirs.each { thisDir ->
                log.debug '*******************************'

                serverContext = thisDir.name
                serverPath = thisDir.listFiles()

                log.debug "Potential serverContext is '$serverContext'"
                log.debug "Directory list is $serverPath"

                serverPath.each { thisPath ->
                    log.debug "thisPath is $thisPath"
                    if (thisPath.isDirectory()) {
                        serverName = thisPath.getName() // last part of the path name
                        ContextCollectionTable.each { logDef ->
// orig vers                            if (logDef.serverPart == serverName) {  // does it match the SAS server subdir name?
                            // changed equation to allow pattern matches
                            if (serverName ==~ logDef.serverPart) {  // does it match the SAS server subdir name?
                                def myfile = canonicalFile(thisPath, logDef.logPart)
                                log.debug "ContextCollection: serverPath=$thisPath serverName=$serverName"
                                log.debug "ContextCollection: serverContext=$serverContext logdir=$myfile"
                                if (verbose) {
                                    msg.info "found: ${logDef.descr}\n        $thisPath"
                                }
                                dirCnt++
                                ProcessLogs(myFile, is_first, thislev, thisPath, serverContext, serverName, logDef)
                                is_first = false
                            }
                        }
                    }
                }  // serverPath.each
            }  //contextDirs.each
        } // lev.each

        // close the end of the JSON map
        myFile.append("\n]\n")
    } catch (Exception e) {
        term.error "A problem was encountered iterating through the levroot hierarchy: $e.message"
        log.error "A problem was encountered iterating through the levroot hierarchy", e
        return exitPgm(RC_GENERAL_ERROR)
    }
    msg.info "Discovered $dirCnt matching log directories"

    return 0
}

void ProcessLogs(File myFile, Boolean is_first, def thislev, File thisPath, String serverContext, String serverName, def logDef) {
    log.debug "process: $myFile, $thisPath, $serverContext, $serverName, $logDef"

    def hostName = thislev.hostFQDN
    def parsedName = (logDef.namePattern.replace('%c%', serverContext)).replace('%h%', hostName)

    try {
        String purgeDays = sysProps['custom.log.purge.days'] ?: (logDef.purgeAfter ?: "0")

        switch(logDef.type) {
            case 'absolute':
                def parsedSrc = logDef.prefixPath.replace('%l%', thislev.levDir) + fs +
                                logDef.serverPart + fs + logDef.logPart

                def parsedDest = (logDef.contextForce ? "${logDef.contextForce}$fs" : "") +
                                 logDef.serverPart + fs + logDef.logPart

                if (!is_first) {
                    myFile << ",\n"
                }
                def builder = new groovy.json.JsonBuilder()
                builder.call ({
                    name "$parsedName"
                    logSource "$parsedSrc"
                    logDest "$parsedDest"
                    filePattern "${logDef.filePattern}"
                    operation "${logDef.operation}"
                    purgeAfter "$purgeDays"
                })
                myFile << groovy.json.JsonOutput.prettyPrint("$builder")
                break

            case 'wildcard':
                def instanceName = thisPath.name
                def parsedSrc = logDef.prefixPath.replace('%l%', thislev.levDir) + fs + instanceName

                def parsedDest = logDef.logDest.replace('%w%', instanceName)

                if (!is_first) {
                    myFile << ",\n"
                }
                def builder = new groovy.json.JsonBuilder()
                builder.call ({
                    name "$parsedName"
                    logSource "$parsedSrc"
                    logDest "$parsedDest"
                    filePattern "${logDef.filePattern}"
                    operation "${logDef.operation}"
                    purgeAfter "$purgeDays"
                })
                myFile << groovy.json.JsonOutput.prettyPrint("$builder")
                break

            case 'context':
//                def logLoc = "%c%/${logDef.serverPart}/${logDef.logPart}"
// modified to allow for patterns in server part
                def logLoc = "%c%/$serverName/${logDef.logPart}"
                def parsedSrc = logDef.prefixPath.replace('%l%', thislev.levDir) + fs + logLoc.replace('%c%', serverContext)
                def parsedDest = logLoc.replace('%c%', serverContext)

                if (!is_first) {
                    myFile << ",\n"
                }
                def builder = new groovy.json.JsonBuilder()
                builder.call ({
                    name "$parsedName"
                    logSource "$parsedSrc"
                    logDest "$parsedDest"
                    filePattern "${logDef.filePattern}"
                    operation "${logDef.operation}"
                    purgeAfter "$purgeDays"
                })
                myFile << groovy.json.JsonOutput.prettyPrint("$builder")
                break

            case 'dotsrv':
                def instanceName = thisPath.getName()
                def dotsrvName = parsedName.replace('%n%', instanceName)

                def parsedSrc = logDef.prefixPath.replace('%l%', thislev.levDir) + fs +
                                instanceName + fs + logDef.logPart

                def parsedDest = (logDef.contextForce ? "${logDef.contextForce}$fs" : "") +
                                 logDef.serverPart + fs + logDef.logPart

                if (!is_first) {
                    myFile << ",\n"
                }
                def builder = new groovy.json.JsonBuilder()
                builder.call ({
                    name "$dotsrvName"
                    logSource "$parsedSrc"
                    logDest "$parsedDest"
                    filePattern "${logDef.filePattern}"
                    operation "${logDef.operation}"
                    purgeAfter "$purgeDays"
                })
                myFile << groovy.json.JsonOutput.prettyPrint("$builder")
                break

            default:
                msg.warn "Unknown log collection type: ${logDef.type}, ignored..."
                break
        } // end switch
    } catch (Exception e) {
        term.error "A problem was encountered processing $thisPath: $e.message"
        log.error "A problem was encountered processing $thisPath", e
        exitPgm(RC_GENERAL_ERROR)
    }
}

def exitPgm(int rc) {
    msg.debug "logDiscovery is exiting with return code $rc"
    System.exit rc
}
